"""
Useful expressions in Keplerian dynamics
"""

import numpy as np
import numpy.linalg as la
from numba import njit

from ._transformations import rotmat1, rotmat3

@njit
def get_inclination(state):
    """Function computes inclination in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
    Returns:
        (float): inclination in radians
    """

    # decompose state to position and velocity vector
    r = state[0:3]
    v = state[3:6]
    # angular momentum
    h = np.cross(r,v)
    # inclination
    inc = np.arccos(h[2] / la.norm(h))
    return inc


@njit
def get_raan(state):
    """Function computes RAAN in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
    Returns:
        (float): RAAN in radians
    """
    
    # decompose state to position and velocity vector
    r = state[0:3]
    v = state[3:6]
    # angular momentum
    h = np.cross(r,v)
    # normal direction of xy-plane
    zdir = np.array([0, 0, 1])
    ndir = np.cross(zdir, h)
    # compute RAAN
    raan = np.arctan2(ndir[1], ndir[0])
    return raan
    
    
@njit
def get_eccentricity(state, mu):
    """Function computes eccentricity vector from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter
    Returns:
        (np.arr): eccentricity vector
    """
    
    # decompose state to position and velocity vector
    r = state[0:3]
    v = state[3:6]
    # angular momentum
    h = np.cross(r,v)
    # normal direction of xy-plane
    #zdir = np.array([0, 0, 1])
    #ndir = np.cross(zdir, h)
    # compute eccentricity vector
    ecc = (1/mu) * np.cross(v,h) - r/la.norm(r)
    return ecc


@njit
def get_omega(state, mu):
    """Function computes argument of periapsis in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter
    Returns:
        (float): argument of periapsis in radians
    """
    
    # decompose state to position and velocity vector
    r = state[0:3]
    v = state[3:6]
    # angular momentum
    h = np.cross(r,v)
    # normal direction of xy-plane
    zdir = np.array([0, 0, 1])
    ndir = np.cross(zdir, h)
    # compute eccentricity vector
    ecc = (1/mu) * np.cross(v,h) - r/la.norm(r)

    # compute argument of periapsis
    if la.norm(ndir)*la.norm(ecc) != 0.0:
        omega = np.arccos( np.dot(ndir,ecc) / (la.norm(ndir)*la.norm(ecc)) )
        if ecc[2] < 0:
            omega = 2*np.pi - omega
    else:
        omega = 0.0
    return omega


@njit
def get_trueanom(state, mu):
    """Function computes argument of periapsis in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter
    Returns:
        (float): true anomaly in radians
    """
    # decompose state to position and velocity vector
    r = state[0:3]
    v = state[3:6]
    # angular momentum
    h = la.norm( np.cross(r,v) )
    # radial velocity
    vr = np.dot(v,r)/la.norm(r)
    theta = np.arctan2(h*vr, h**2/la.norm(r) - mu)
    return theta


@njit
def get_semiMajorAxis(state, mu):
    """Function computes semi major axis of keplrian orbit
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter
    Returns:
        (float): semi-major axis
    """
    # decompose state to position and velocity vector
    r = state[0:3]
    v = state[3:6]
    # angular momentum
    h = la.norm( np.cross(r,v) )
    # eccentricity
    e = la.norm( get_eccentricity(state, mu) )
    # semi-major axis
    a = h**2 / (mu*(1 - e**2))
    return a


@njit
def get_period(state, mu):
    """Function computes period of keplerian orbit
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter
        
    Returns:
        (float): period
    """
    a = get_semiMajorAxis(state, mu)
    if a < 0:
        period = -1
    else:
        period = 2*np.pi*np.sqrt(a**3/mu)
    return period


@njit
def sv2kep(state, mu):
    """Convert Cartesian states to Keplerian elements
    Args:
        state (np.array): cartesian states
        mu (float): two-body mass parameter
    
    Returns:
        (np.array): elements, in order [sma, ecc, inc, raan, aop, ta]
    """
    # convert to orbital elements
    sma = get_semiMajorAxis(state, mu)
    ecc = la.norm(get_eccentricity(state, mu))
    inc = get_inclination(state)
    raan = get_raan(state)
    aop = get_omega(state, mu)
    ta = get_trueanom(state, mu)
    return np.array([sma, ecc, inc, raan, aop, ta])


@njit
def kep2sv(kep, mu):
    """Convert Keplerian elements to Cartesian states
    Args:
        kep (np.array): elements, in order [sma, ecc, inc, raan, aop, ta]
        mu (float): two-body mass parameter
    
    Returns:
        (np.array): cartesian states
    """
    # unpack elements
    sma, ecc, inc, raan, aop, ta = kep
    # construct state in perifocal frame
    h = np.sqrt(sma*mu*(1 - ecc**2))
    rPF = h**2 / (mu*(1 + ecc*np.cos(ta))) * np.array([np.cos(ta), np.sin(ta), 0.0])
    vPF = mu/h * np.array([-np.sin(ta), ecc + np.cos(ta), 0.0])
    # convert to inertial frame
    if (aop != 0.0):
        r1 = np.dot(rotmat3(-aop), rPF)
        v1 = np.dot(rotmat3(-aop), vPF)
    else:
        r1 = rPF
        v1 = vPF

    if (inc != 0.0):
        r2 = np.dot(rotmat1(-inc), r1)
        v2 = np.dot(rotmat1(-inc), v1)
    else:
        r2 = r1
        v2 = v1
    
    if (raan != 0.0):
        rI = np.dot(rotmat3(-raan), r2)
        vI = np.dot(rotmat3(-raan), v2)
    else:
        rI = r2
        vI = v2
    return np.concatenate((rI, vI))


@njit
def get_orbit_coordinates(oe_kep,mu,steps=200):
    # unpack elements
    sma, ecc, inc, raan, aop, _ = oe_kep
    tas = np.linspace(0.0, 2*np.pi, steps)
    coord = np.zeros((6,steps))
    for idx in range(steps):
        coord[:,idx] = kep2sv(np.array([sma, ecc, inc, raan, aop, tas[idx]]), mu)
    return coord


@njit
def kep2mee(oe_kep):
    """Convert Keplerian elements to MEE"""
    # unpack
    a,e,i,raan,om,ta = oe_kep
    # compute MEEs
    p = a*(1-e**2)
    f = e*np.cos(raan+om)
    g = e*np.sin(raan+om)
    h = np.tan(i/2)*np.cos(raan)
    k = np.tan(i/2)*np.sin(raan)
    l = raan + om + ta
    return np.array([p,f,g,h,k,l])


@njit
def mee2kep(oe_mee):
    """Convert MEE to Keplerian elements"""
    # unpack 
    p,f,g,h,k,l = oe_mee
    # compute Keplerian elements
    a = p/(1-f**2-g**2)
    e = np.sqrt(f**2 + g**2)
    i = np.arctan2(2*np.sqrt(h**2+k**2), 1-h**2-k**2)
    raan = np.arctan2(k,h)
    om = np.arctan2(g*h-f*k, f*h+g*k)
    ta = l - raan - om
    return np.array([a,e,i,raan,om,ta])


@njit
def mee_with_a2mee(oe_mee_with_a):
    # unpack 
    a,f,g,h,k,l = oe_mee_with_a
    p = a * (1-f**2-g**2)
    return np.array([p, f,g,h,k,l])


def mee_with_a2kep(oe_mee_with_a):
    return mee2kep(mee_with_a2mee(oe_mee_with_a))


@njit
def mee2mee_with_a(oe_mee):
    """Convert MEE to MEE with SMA"""
    # unpack 
    p,f,g,h,k,l = oe_mee
    a = p/(1-f**2-g**2)
    return np.array([a, f,g,h,k,l])


@njit
def kep2mee_with_a(oe_kep):
    """Get targeting element set used by Q-law when using MEE"""
    # unpack
    a,e,i,raan,om,ta = oe_kep
    # compute MEEs
    #p = a*(1-e**2)
    f = e*np.cos(raan+om)
    g = e*np.sin(raan+om)
    h = np.tan(i/2)*np.cos(raan)
    k = np.tan(i/2)*np.sin(raan)
    l = raan + om + ta
    return np.array([a,f,g,h,k,l])


#@njit
def mee_with_a2sv(mee_with_a, mu):
    """Convert MEE with SMA to Cartesian states"""
    a,f,g,h,k,l = mee_with_a
    mee = np.array([a * (1 - f**2 - g**2), f,g,h,k,l])
    kep = mee2kep(mee)
    return kep2sv(kep, mu)


@njit
def ta2ea(ta,ecc):
    """Convert true anomaly to eccentric anomaly"""
    ea = 2*np.arctan(np.sqrt((1-ecc)/(1+ecc))*np.tan(ta/2))
    return ea


@njit
def mee2ea(mee):
    ecc = np.sqrt(mee[1]**2 + mee[2]**2)
    ta = mee[5] - np.arctan2(mee[4],mee[3])
    return ta2ea(ta,ecc)


"""
Lyapunov feedback control
"""

import numpy as np
import numpy.linalg as la
from numba import njit

@njit
def _u_to_thrust_angles(u_float):
    """Convert unscaled vector u_float to unit-vector u & thrust angles alpha and beta"""
    u_float_norm = np.sqrt( u_float[0]**2 + u_float[1]**2 + u_float[2]**2 )
    u = u_float/u_float_norm
    alpha = np.arctan2(-u[0],-u[1])
    beta = np.arctan(-u[2]/np.sqrt(u[0]**2 + u[1]**2))
    return u, alpha, beta


def lyapunov_control_angles(
        fun_lyapunov_control,
        mu, 
        f, 
        oe, 
        oeT, 
        rpmin, m_petro, n_petro, 
        r_petro, b_petro, k_petro, 
        wp, 
        woe,
    ):
    """Compute thrust angles from Lyapunov feedback control law
    
    Args:
        fun_lyapunov_control (callable): symbolic-generated Lyapunov control law
        mu (float): gravitational parameter
        f (float): thrust-acceleration
        oe (np.array): current osculating elements
        oeT (np.array): target osculating elements
        rpmin (float): minimum periapsis
        k_petro (float): scalar factor k on minimum periapsis
        m_petro (float): scalar factor m to prevent non-convergence, default is 3.0
        n_petro (float): scalar factor n to prevent non-convergence, default is 4.0
        r_petro (float): scalar factor r to prevent non-convergence, default is 2.0
        b_petro (float): scalar factor b for dot(omega)_xx, default is 0.01
        wp (float): penalty scalar on minimum periapsis, default is 1.0
        woe (np.array): weight on each osculating element
    
    Returns:
        (tuple): alpha, beta, vector u, list of columns of psi
    """
    # compute u direction
    u_raw, psi = fun_lyapunov_control(
        mu, 
        f, 
        oe, 
        oeT, 
        rpmin, m_petro, n_petro, 
        r_petro, b_petro, k_petro, 
        wp, 
        woe
    )

    # compute thrust angles
    u_float = np.array([float(el) for el in u_raw])
    u, alpha, beta = _u_to_thrust_angles(u_float)
    return alpha, beta, u, psi


"""
integration
"""

import numpy as np
from numba import njit

@njit
def rk4(rhs, t, h, y, p):
    """Perform single-step Runge-Kutta 4th order
    
    Args:
        rhs (callable): ODE right-hand side expressions
        t (float): current time
        h (float): time-step
        y (np.array): current state-vector
        p (Real or tuple): additional parameters passed to `rhs`

    Returns:
        (np.array); state-vector at time t+h
    """
    k1 = rhs(t, y, p)
    k2 = rhs(t + 0.5 * h, y + 0.5 * k1, p)
    k3 = rhs(t + 0.5 * h, y + 0.5 * k2, p)
    k4 = rhs(t + h, y + k3, p)
    return y + (h / 6.0)*(k1 + 2*k2 + 2*k3 + k4)


@njit
def rkf45(rhs, t, h, y, p, tol=1e-6):
    """Perform single-step Runge-Kutta 4th order with 5th order step correction
    
    Args:
        rhs (callable): ODE right-hand side expressions
        t (float): current time
        h (float): time-step
        y (np.array): current state-vector
        p (Real or tuple): additional parameters passed to `rhs`
        tol (float): tolerance

    Returns:
        (np.array); state-vector at time t+h
    """
    k1 = h * rhs(t, y, p)
    k2 = h * rhs(t + 1/4 * h,   y + 1/4*h * k1, p)
    k3 = h * rhs(t + 3/8 * h,   y + 3/32*h*k1 + 9/32*h*k2, p)
    k4 = h * rhs(t + 12/13 * h, y + 1932/2197*h*k1 - 7200/2197*h*k2 + 7296/2197*h*k3, p)
    k5 = h * rhs(t + h,         y + 439/216*h*k1 - 8*h*k2 + 3680/513*h*k3 - 845/4104*h*k4, p)
    k6 = h * rhs(t + 1/2*h,     y - 8/27*h*k1 + 2*h*k2 - 3544/2565*h*k3 + 1859/4104*h*k4 - 11/40*h*k5, p)
    y_next = y + 25/216*k1 + 1408/2565*k3 + 2197/4101*k4 - 1/5*k5
    z_next = y + 16/135*k1 + 6656/12825*k3 + 28561/56430*k4 - 9/50*k5 + 2/55*k6
    h_next = h*0.84*(tol/np.linalg.norm(z_next - y_next))**(1/4)
    return y_next, h_next


"""
Check convergence
"""

import copy
import numpy as np
from numba import njit


@njit
def check_convergence(oe, oeT, woe, tol_oe):
    """Check convergence between oe and oeT"""
    check_array = np.zeros(5,) # np.array([1 if el == 0.0 else 0 for el in woe])
    doe = np.abs(oeT[0:5] - oe[0:5])  # FIXME is this ok? or should we be careful for angles?
    check_array = doe < tol_oe

    if np.sum(check_array)==5:
        return True
    else:
        return False


@njit
def elements_safety(oe, oe_min, oe_max):
    """Ensure osculating elements stays within bounds
    
    Args:
        oe (np.array): current osculating elements
        oe_min (np.array): minimum values for each osculating element
        oe_max (np.array): maximum values for each osculating element

    Returns:
        (np.array): "cleaned" osculating elements
    """
    oe_clean = oe
    for idx in range(5):
        oe_clean[idx] = min(max(oe_min[idx], oe[idx]), oe_max[idx])
    return oe_clean


"""
Symbolic derivation for feedback control law
"""

"""
sympy dev
"""

import numpy as np
import sympy as sym
from sympy import lambdify


def symbolic_qlaw_keplerian(cse = True):
    """Generate symbolic function for Keplerian Q-Law"""
    # define parameters
    rpmin, k_petro, m_petro, n_petro, r_petro, b_petro = sym.symbols("rpmin k_petro m_petro n_petro r_petro b_petro")
    wp = sym.symbols("Wp")
    mu = sym.symbols("mu")
    f = sym.symbols("f")

    # orbital elements
    a, e, i, ra, om, ta = sym.symbols("a e i Omega omega theta")
    oe = [a,e,i,ra,om,ta]

    # targeted orbital elements
    aT, eT, iT, raT, omT = sym.symbols("a_T e_T i_T Omega_T omega_T")
    oeT = [aT, eT, iT, raT, omT]

    # weights on orbital elements
    wa, we, wi, wra, wom = sym.symbols("w_a w_e w_i w_Omega w_omega")
    woe = [wa, we, wi, wra, wom]


    def quotient(mu, f, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe):
        # unpack elements
        a,e,i,ra,om,ta = oe
        aT, eT, iT, raT, omT = oeT
        rp = a*(1-e)
        p = a*(1 - e**2)
        h = sym.sqrt(a*mu*(1-e**2))

        # sin and cos terms
        cosTA = sym.cos(ta)
        sinTA = sym.sin(ta)
        cosi  = sym.cos(i)
        sini  = sym.sin(i)
        cosom = sym.cos(om)
        sinom = sym.sin(om)

        # -------- compute quotient Q -------- #
        doe = [
            a-aT, e-eT, i-iT, 
            sym.acos(sym.cos(ra - raT)),
            sym.acos(sym.cos(om - omT)),
            #sym.acos(sym.cos(ta - taT)),
        ]

        # compute RAdot and omdot
        radot = p*f/(h*sini*(sym.sqrt(1 - e**2*cosom**2) - e*sym.sqrt((sinom**2))))  # FIXME
        cos_ta_xx_sqrt = sym.sqrt((1/4)*((1-e**2)/e**3)**2 + 1/27)
        ta_xx_term1 = ( (1-e**2) / (2*e**3) + cos_ta_xx_sqrt)**(1/3)
        ta_xx_term2 = (-(1-e**2) / (2*e**3) + cos_ta_xx_sqrt)**(1/3)
        cos_ta_xx = ta_xx_term1 - ta_xx_term2 - 1/e
        r_xx = p / (1 + e*cos_ta_xx)
        omdot_i = f/(e*h)*sym.sqrt(
        p**2*cos_ta_xx**2 + (p+r_xx)**2*(1-cos_ta_xx**2)
        )
        omdot_o = radot*sym.sqrt(cosi**2)
        omdot   = (omdot_i + b_petro*omdot_o)/(1+b_petro)

        # compute oedot
        oedot = [
            2*f*sym.sqrt(a**3 * (1+e)/(mu*(1-e))),  # adot
            2*p*f/h,                                # edot
            p*f/(h*(sym.sqrt(1 - e**2*sinom**2) - e*sym.sqrt(cosom**2))),  # idot
            radot,
            omdot,
        ]
        
        # compute periapsis radius constraint P
        p_rp = sym.exp(k_petro*(1.0 - rp/rpmin))
        # compute scaling for each element Soe
        soe = [(1 + ((a-aT)/(m_petro*aT))**n_petro)**(1/r_petro), 1.0, 1.0, 1.0, 1.0, 1.0]
        # compute quotient Q
        #sum_term = 0
        #for idx in range(5):
        #    sum_term += woe[idx]*soe[idx]*(doe[idx]/oedot[idx])**2
        q = (1 + wp*p_rp) * (
            woe[0]*soe[0]*(doe[0]/oedot[0])**2 +
            woe[1]*soe[1]*(doe[1]/oedot[1])**2 + 
            woe[2]*soe[2]*(doe[2]/oedot[2])**2 +
            woe[3]*soe[3]*(doe[3]/oedot[3])**2 +
            woe[4]*soe[4]*(doe[4]/oedot[4])**2 
        )

        # -------- compute Gauss differential equation terms -------- #
        r = h**2/(mu*(1+e*cosTA))
        # let psi be column major!
        psi = [
            # multiplies f_r
            [
                2*a**2/h*e*sinTA,
                p/h*sinTA,
                0.0,
                0.0,
                -p/(e*h)*cosTA
            ],
            # multiplies f_theta
            [
                2*a**2/h * p/r,
                ((p+r)*cosTA + r*e)/h,
                0.0,
                0.0,
                (p+r)*sinTA/(e*h)
            ],
            # multiplies f_h
            [
                0.0,
                0.0,
                r*sym.cos(ta+om)/h,
                r*sym.sin(ta+om)/(h*sini),
                -r*sym.sin(ta+om)*cosi/(h*sini),
            ]
        ]

        # -------- Apply Lyapunov descent direction -------- #
        # dqdoe = [
        #     sym.diff(q, oe[0]),
        #     sym.diff(q, oe[1]),
        #     sym.diff(q, oe[2]),
        #     sym.diff(q, oe[3]),
        #     sym.diff(q, oe[4]),
        # ]
        dqdoe0 = sym.diff(q, oe[0])   # a
        dqdoe1 = sym.diff(q, oe[1])   # e
        dqdoe2 = sym.diff(q, oe[2])   # i 
        dqdoe3 = sym.diff(q, oe[3])   # RA
        dqdoe4 = sym.diff(q, oe[4])   # om  --- FIXME

        ux = (psi[0][0]*dqdoe0 + psi[0][1]*dqdoe1 + psi[0][2]*dqdoe2 + psi[0][3]*dqdoe3 + psi[0][4]*dqdoe4)
        uy = (psi[1][0]*dqdoe0 + psi[1][1]*dqdoe1 + psi[1][2]*dqdoe2 + psi[1][3]*dqdoe3 + psi[1][4]*dqdoe4)
        uz = (psi[2][0]*dqdoe0 + psi[2][1]*dqdoe1 + psi[2][2]*dqdoe2 + psi[2][3]*dqdoe3 + psi[2][4]*dqdoe4)

        # compute qdot
        #qdot = ux + uy + uz
        dqdt = dqdoe0*(psi[0][0]*ux + psi[1][0]*uy + psi[2][0]*uz) +\
               dqdoe1*(psi[0][1]*ux + psi[1][1]*uy + psi[2][1]*uz) +\
               dqdoe2*(psi[0][2]*ux + psi[1][2]*uy + psi[2][2]*uz) +\
               dqdoe3*(psi[0][3]*ux + psi[1][3]*uy + psi[2][3]*uz) +\
               dqdoe4*(psi[0][4]*ux + psi[1][4]*uy + psi[2][4]*uz)

        fun_lyapunov_control = lambdify(
            [mu, f, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            [[ux,uy,uz], psi], 
            "numpy",
            cse = cse,
        )

        fun_eval_psi = lambdify(
            [mu, f, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            psi, 
            "numpy",
            cse = cse,
        )

        fun_eval_dqdoe = lambdify(
            [mu, f, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            [dqdoe0,dqdoe1,dqdoe2,dqdoe3,dqdoe4], 
            "numpy",
            cse = cse,
        )

        fun_eval_dqdt = lambdify(
            [mu, f, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            dqdt, 
            "numpy",
            cse = cse,
        )
        #sym.diff(q, oe[2]), sym.diff(q, oe[4])
        return fun_lyapunov_control, fun_eval_psi, fun_eval_dqdoe, fun_eval_dqdt

    # create function
    #print("Generating Keplerian lyapunov control funcion with sympy")
    fun_lyapunov_control, fun_eval_psi, fun_eval_dqdoe, fun_eval_dqdt = quotient(mu, f, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe)
    return fun_lyapunov_control, fun_eval_psi, fun_eval_dqdoe, fun_eval_dqdt



def symbolic_qlaw_mee_with_a(cse = True):
    """Generate symbolic function for Keplerian Q-Law"""
    # define parameters
    rpmin, k_petro, m_petro, n_petro, r_petro, b_petro = sym.symbols("rpmin k_petro m_petro n_petro r_petro b_petro")
    wp = sym.symbols("Wp")
    mu = sym.symbols("mu")
    accel = sym.symbols("accel")

    # orbital elements
    a, f, g, h, k, l = sym.symbols("a f g h k l")
    oe = [a,f,g,h,k,l]

    # targeted orbital elements
    aT, fT, gT, hT, kT, lT = sym.symbols("a_T f_T g_T h_T k_T l_T")
    oeT = [aT, fT, gT, hT, kT]

    # weights on orbital elements
    wa, wf, wg, wh, wk = sym.symbols("w_a w_f w_g w_h w_k")
    woe = [wa, wf, wg, wh, wk]


    def angle_difference(phi1, phi2):
        return sym.acos(sym.cos(phi1 - phi2))

    def quotient(mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe):
        # unpack elements
        a, f, g, h, k, l = oe
        aT, fT, gT, hT, kT = oeT

        # compute semi-parameter p, periapsis rp
        p = a * (1 - f**2 - g**2)
        e = sym.sqrt(f**2 + g**2)
        ang_mom = sym.sqrt(a*mu*(1-e**2))
        rp = a*(1 - e)
        ta = l - sym.atan(g/f)
        r = ang_mom**2/(mu*(1+e*sym.cos(ta)))
        s_squared = 1 + h**2 + k**2

        # -------- compute quotient Q -------- #
        doe = [
            a-aT, 
            f-fT, 
            g-gT, 
            h-hT, 
            k-kT,
        ]

        # compute oedot
        sqrt_pmu = sym.sqrt(p/mu)
        adot_xx = 2*accel*a*sym.sqrt(a/mu) * sym.sqrt((1 + sym.sqrt(f**2 + g**2)) / (1 - sym.sqrt(f**2 + g**2)))
        fdot_xx = 2*accel*sqrt_pmu
        gdot_xx = 2*accel*sqrt_pmu
        hdot_xx = 0.5*accel*sqrt_pmu * s_squared/(sym.sqrt(1 - g**2) + f)
        kdot_xx = 0.5*accel*sqrt_pmu * s_squared/(sym.sqrt(1 - f**2) + g)
        oedot = [
            adot_xx,
            fdot_xx,
            gdot_xx,
            hdot_xx,
            kdot_xx,
        ]
        
        # compute periapsis radius constraint P
        penalty_rp = sym.exp(k_petro*(1.0 - rp/rpmin))
        # compute scaling for each element Soe
        soe = [
            (1 + (sym.sqrt((a-aT)**2)/(m_petro*aT))**n_petro)**(1/r_petro), 
            1.0, 1.0, 1.0, 1.0
        ]
        # compute quotient Q
        q = (1 + wp*penalty_rp) * (
            woe[0]*soe[0]*(doe[0]/oedot[0])**2 +
            woe[1]*soe[1]*(doe[1]/oedot[1])**2 + 
            woe[2]*soe[2]*(doe[2]/oedot[2])**2 +
            woe[3]*soe[3]*(doe[3]/oedot[3])**2 +
            woe[4]*soe[4]*(doe[4]/oedot[4])**2 
        )

        # -------- compute Gauss differential equation terms -------- #
        cosL = sym.cos(l)
        sinL = sym.sin(l)
        w = 1 + f*cosL + g*sinL
        # let psi be column major!
        psi = [
            # multiplies f_r
            [
                2*a**2/ang_mom*e*sym.sin(ta), # this is for a!! #0.0,
                sqrt_pmu* sinL,
                -sqrt_pmu* cosL,
                0.0,
                0.0,
            ],
            # multiplies f_theta
            [
                2*a**2/ang_mom * p/r,  # this is for a!!! #sqrt_pmu*2*p/w,
                sqrt_pmu/w * ((w+1)*cosL + f),
                sqrt_pmu/w * ((w+1)*sinL + g),
                0.0,
                0.0,
            ],
            # multiplies f_h
            [
                0.0,  # this is for a!!!
                sqrt_pmu/w* (-g*(h*sinL - k*cosL)),
                sqrt_pmu/w* ( f*(h*sinL - k*cosL)),
                sqrt_pmu/w* 0.5*(1 + h**2 + k**2)*cosL,
                sqrt_pmu/w* 0.5*(1 + h**2 + k**2)*sinL,
            ]
        ]

        # -------- Apply Lyapunov descent direction -------- #
        # dqdoe = [
        #     sym.diff(q, oe[0]),
        #     sym.diff(q, oe[1]),
        #     sym.diff(q, oe[2]),
        #     sym.diff(q, oe[3]),
        #     sym.diff(q, oe[4]),
        # ]
        dqdoe0 = sym.diff(q, oe[0])   # a
        dqdoe1 = sym.diff(q, oe[1])   # f
        dqdoe2 = sym.diff(q, oe[2])   # g 
        dqdoe3 = sym.diff(q, oe[3])   # h
        dqdoe4 = sym.diff(q, oe[4])   # k

        # # compute thrust vector components
        # u_unscaled = [
        #     psi[0][0]*dqdoe0 + psi[0][1]*dqdoe1 + psi[0][2]*dqdoe2 + psi[0][3]*dqdoe3 + psi[0][4]*dqdoe4,
        #     psi[1][0]*dqdoe0 + psi[1][1]*dqdoe1 + psi[1][2]*dqdoe2 + psi[1][3]*dqdoe3 + psi[1][4]*dqdoe4,
        #     psi[2][0]*dqdoe0 + psi[2][1]*dqdoe1 + psi[2][2]*dqdoe2 + psi[2][3]*dqdoe3 + psi[2][4]*dqdoe4
        # ]
        # u_unscaled_norm = sym.sqrt(u_unscaled[0]**2 + u_unscaled[1]**2 + u_unscaled[2]**2)
        # ux = u_unscaled[0]/u_unscaled_norm
        # uy = u_unscaled[0]/u_unscaled_norm
        # uz = u_unscaled[0]/u_unscaled_norm

        # # compute thrust angles
        # alpha = sym.atan2(-ux,-uy)
        # beta = sym.atan(-uz/sym.sqrt(ux**2 + uy**2))

        # # compute thrust vector
        # # thrust_vec = [
        # #     accel* sym.cos(beta)*sym.sin(alpha),
        # #     accel* sym.cos(beta)*sym.cos(alpha),
        # #     accel* sym.sin(beta),
        # # ]

        # fun_lyapunov_control = lambdify(
        #     [
        #         mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, 
        #         b_petro, k_petro, wp, woe
        #     ], 
        #     [psi, qdot, alpha, beta], 
        #     "numpy",
        # )

        ux = (psi[0][0]*dqdoe0 + psi[0][1]*dqdoe1 + psi[0][2]*dqdoe2 + psi[0][3]*dqdoe3 + psi[0][4]*dqdoe4)
        uy = (psi[1][0]*dqdoe0 + psi[1][1]*dqdoe1 + psi[1][2]*dqdoe2 + psi[1][3]*dqdoe3 + psi[1][4]*dqdoe4)
        uz = (psi[2][0]*dqdoe0 + psi[2][1]*dqdoe1 + psi[2][2]*dqdoe2 + psi[2][3]*dqdoe3 + psi[2][4]*dqdoe4)

        dqdt = dqdoe0*(psi[0][0]*ux + psi[1][0]*uy + psi[2][0]*uz) +\
               dqdoe1*(psi[0][1]*ux + psi[1][1]*uy + psi[2][1]*uz) +\
               dqdoe2*(psi[0][2]*ux + psi[1][2]*uy + psi[2][2]*uz) +\
               dqdoe3*(psi[0][3]*ux + psi[1][3]*uy + psi[2][3]*uz) +\
               dqdoe4*(psi[0][4]*ux + psi[1][4]*uy + psi[2][4]*uz)

        fun_lyapunov_control = lambdify(
            [mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            [[ux, uy, uz], psi], 
            "numpy",
            cse = cse,
        )

        fun_eval_psi = lambdify(
            [mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            psi, 
            "numpy",
            cse = cse,
        )

        fun_eval_dqdoe = lambdify(
            [
                mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, 
                b_petro, k_petro, wp, woe
            ], 
            [dqdoe0,dqdoe1,dqdoe2,dqdoe3,dqdoe4], 
            "numpy",
            cse = cse,
        )

        fun_eval_dqdt = lambdify(
            [mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe], 
            dqdt, 
            "numpy",
            cse = cse,
        )
        return fun_lyapunov_control, fun_eval_psi, fun_eval_dqdoe, fun_eval_dqdt

    # create function
    #print("Generating MEE-SMA lyapunov control funcion with sympy")
    fun_lyapunov_control, fun_eval_psi, fun_eval_dqdoe, fun_eval_dqdt = quotient(mu, accel, oe, oeT, rpmin, m_petro, n_petro, r_petro, b_petro, k_petro, wp, woe)
    return fun_lyapunov_control, fun_eval_psi, fun_eval_dqdoe, fun_eval_dqdt



"""
init file for module
"""

__copyright__    = 'Copyright (C) 2023 Yuri Shimane'
__version__      = '0.1.1'
__license__      = 'MIT License'
__author__       = 'Yuri Shimane'
__author_email__ = 'yuri.shimane@gatech.edu'
__url__          = 'https://github.com/Yuricst/pyqlaw'


from ._symbolic import (
    symbolic_qlaw_keplerian,
    symbolic_qlaw_mee_with_a,
)
from ._eom import (
    eom_kep_gauss, 
    eom_mee_with_a_gauss,
)
from ._integrate import (
    rk4, 
    rkf45,
)
from ._lyapunov import lyapunov_control_angles
from ._elements import (
    sv2kep, 
    kep2sv,
    kep2mee,
    mee2kep,
    kep2mee_with_a,
    mee_with_a2sv,
    ta2ea,
    mee2ea,
)
from ._convergence import check_convergence
from ._qlaw import QLaw
from ._plot_helper import (
    get_sphere_coordinates,
    plot_sphere_wireframe,
    get_ellipsoid_coordinates,
    plot_ellipsoid_wireframe,
    set_equal_axis,
)


"""
Test for passing eta's as functions
"""


import numpy as np
from numpy.random import rand
import matplotlib.pyplot as plt
import time 

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))
import pyqlaw


def plot_eta_func():
    """View eta function"""
    # initial and final elements: [a,e,i,RAAN,omega,ta]
    LU = 42164.0
    GM_EARTH = 398600.44
    VU = np.sqrt(GM_EARTH/LU)
    TU = LU/VU

    rp_gto = 200 + 6378
    ra_gto = 35786 + 6378
    sma_gto = (rp_gto + ra_gto)/(2*LU)
    ecc_gto = (ra_gto - rp_gto)/(ra_gto + rp_gto)
    KEP0 = [sma_gto,ecc_gto,np.deg2rad(23),0,0,0]
    KEPF = [1,0,np.deg2rad(3),0,0,0]
    oe0 = pyqlaw.kep2mee_with_a(np.array(KEP0))
    oeT = pyqlaw.kep2mee_with_a(np.array(KEPF))
    woe = [3.0, 1.0, 1.0, 1.0, 1.0]

    # spacecraft parameters
    MU = 1500
    tmax_si = 1.0    # 1 N
    isp_si  = 1500   # seconds
    mdot_si = tmax_si/(isp_si*9.81)  # kg/s

    # non-dimensional quantities
    mass0 = 1.0
    tmax = tmax_si * (1/MU)*(TU**2/(1e3*LU))
    mdot = np.abs(mdot_si) *(TU/MU)
    tf_max = 10000.0
    t_step = np.deg2rad(15)
    
    # battery levels
    battery_initial = 3000*3600/TU            # Wh --> Ws --> W.TU
    battery_dod = 500*3600/TU
    battery_capacity = (battery_dod,battery_initial)

    def eta_r(t,oe,mass,battery):
        return min(400/(4*battery - battery_capacity[0]), 1)

    batteries = np.linspace(battery_capacity[0], battery_capacity[1], 500)    
    eta_r_eval = np.array([eta_r(0,0,0,b) for b in batteries])

    fig, ax = plt.subplots(1,1,figsize=(6,4))
    ax.plot(batteries*TU/3600, eta_r_eval)
    ax.set(xlim=np.array(battery_capacity)*TU/3600,
           ylabel='eta_r_min', xlabel='battery level, W.hr')
    return
    

def test_object():
    tstart = time.time()

    # initial and final elements: [a,e,i,RAAN,omega,ta]
    LU = 42164.0
    GM_EARTH = 398600.44
    VU = np.sqrt(GM_EARTH/LU)
    TU = LU/VU

    rp_gto = 200 + 6378
    ra_gto = 35786 + 6378
    sma_gto = (rp_gto + ra_gto)/(2*LU)
    ecc_gto = (ra_gto - rp_gto)/(ra_gto + rp_gto)
    KEP0 = [sma_gto,ecc_gto,np.deg2rad(23),0,0,0]
    KEPF = [1,0,np.deg2rad(3),0,0,0]
    oe0 = pyqlaw.kep2mee_with_a(np.array(KEP0))
    oeT = pyqlaw.kep2mee_with_a(np.array(KEPF))
    woe = [3.0, 1.0, 1.0, 1.0, 1.0]

    # spacecraft parameters
    MU = 1500
    tmax_si = 1.0    # 1 N
    isp_si  = 1500   # seconds
    mdot_si = tmax_si/(isp_si*9.81)  # kg/s

    # non-dimensional quantities
    mass0 = 1.0
    tmax = tmax_si * (1/MU)*(TU**2/(1e3*LU))
    mdot = np.abs(mdot_si) *(TU/MU)
    tf_max = 10000.0
    t_step = np.deg2rad(15)
    
    # battery levels
    battery_initial = 3000*3600/TU            # Wh --> Ws --> W.TU
    battery_dod = 500*3600/TU
    battery_capacity = (battery_dod,battery_initial)
    charge_rate = 1500          # W
    discharge_rate = 500        # W
    battery_charge_discharge_rate = (charge_rate, discharge_rate)
    require_full_recharge = True
    print(f"battery_capacity = {battery_capacity}")

    # construct problem
    prob = pyqlaw.QLaw(
        integrator="rk4", 
        elements_type="mee_with_a",
        verbosity=2,
        print_frequency=500,
        use_sundman = True,
    )

    # set problem
    prob.set_problem(oe0, oeT, mass0, tmax, mdot, tf_max, t_step,
        battery_initial = battery_initial,
        battery_capacity = battery_capacity,
        battery_charge_discharge_rate = battery_charge_discharge_rate,
        require_full_recharge = require_full_recharge,
        woe = woe)
    prob.pretty()

    # efficiency functiond
    def eta_a(t,oe,mass,battery):
        return 0.0    # not using absolute efficiency threshold
    
    def eta_r(t,oe,mass,battery):
        return min(400/(4*battery - battery_capacity[0]), 1)

    # solve
    tstart_solve = time.time()
    prob.solve(eta_a=eta_a, eta_r=eta_r)
    prob.pretty_results()
    tend = time.time()
    print(f"Simulation took {tend-tstart:4.4f} seconds")
    print(f"prob.solve took {tend-tstart_solve:4.4f} seconds")

    # plot
    fig1, ax1 = prob.plot_elements_history(to_keplerian=True, TU=TU/86400, time_unit_name="day")
    fig2, ax2 = prob.plot_trajectory_3d(sphere_radius=0.1)
    fig3, ax3 = prob.plot_controls()
    fig4, ax4 = prob.plot_battery_history(TU=TU/86400, BU=TU/3600,
        time_unit_name="day", battery_unit_name="Wh")
    fig5, ax5 = prob.plot_efficiency(TU=TU/86400, time_unit_name="day")
    assert prob.converge == True


if __name__=="__main__":
    plot_eta_func()
    figs = test_object()
    plt.show()
    print("Done!")


"""
Helper functions for plotting
"""

import numpy as np
import matplotlib.pyplot as plt

def get_sphere_coordinates(radius, center=None):
    """Get x,y,z coordinates for sphere

    Args:
        radius (float): sphere radius
        center (list): x,y,z coordinates of center; if None, set to [0.0, 0.0, 0.0]
    
    Returns:
        (tuple): x, y, z coordinates of sphere
    """
    # check if center is provided
    if center is None:
        center = [0.0, 0.0, 0.0]
    # construct reference sphere
    u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
    #u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
    x_sphere = center[0] + radius*np.cos(u)*np.sin(v)
    y_sphere = center[1] + radius*np.sin(u)*np.sin(v)
    z_sphere = center[2] + radius*np.cos(v)
    return x_sphere, y_sphere, z_sphere


def plot_sphere_wireframe(ax, radius, center=None, color="k", linewidth=0.5):
    """Plot sphere wireframe
    
    Args:
        ax (Axes3DSubplot): matplotlib 3D axis, created by `ax = fig.add_subplot(projection='3d')`
        radius (float): radius
        center (list): x,y,z coordinates of center, if None set to [0.0, 0.0, 0.0]
        color (str): color of wireframe
        linewidth (float): linewidth of wireframe
    """
    x_sphere, y_sphere, z_sphere = get_sphere_coordinates(radius, center)
    ax.plot_wireframe(x_sphere, y_sphere, z_sphere, color=color, linewidth=linewidth)
    return


def get_ellipsoid_coordinates(rx, ry, rz, center=None, n=60):
    """Get x,y,z coordinates for ellipsoid

    Args:
        rx (float): radius in x-direction
        ry (float): radius in y-direction
        rz (float): radius in z-direction
        center (list): x,y,z coordinates of center; if None, set to [0.0, 0.0, 0.0]
        n (int): number of points to be used in each mesh direction

    Returns:
        (tuple): x, y, z coordinates of ellipsoid
    """
    # check if center is provided
    if center is None:
        center = [0.0, 0.0, 0.0]

    # grid setup
    u = np.linspace(0, 2 * np.pi, n)
    v = np.linspace(0, np.pi, n)

    # Cartesian coordinates of ellipsoid
    x_el = center[0] + rx * np.outer(np.cos(u), np.sin(v))
    y_el = center[1] + ry * np.outer(np.sin(u), np.sin(v))
    z_el = center[2] +rz * np.outer(np.ones_like(u), np.cos(v))
    return x_el, y_el, z_el


def plot_ellipsoid_wireframe(ax, rx, ry, rz, center=None, color="k", linewidth=0.5, n=60):
    """Plot ellipsoid wireframe

    Args:
        ax (Axes3DSubplot): matplotlib 3D axis, created by `ax = fig.add_subplot(projection='3d')`
        rx (float): radius in x-direction
        ry (float): radius in y-direction
        rz (float): radius in z-direction
        center (list): x,y,z coordinates of center, if None set to [0.0, 0.0, 0.0]
        color (str): color of wireframe
        linewidth (float): linewidth of wireframe
        n (int): number of points to be used in each mesh direction
    """
    x_el, y_el, z_el = get_ellipsoid_coordinates(rx, ry, rz, center=center, n=n)
    ax.plot_wireframe(x_el, y_el, z_el, color=color, linewidth=linewidth)
    return


def set_equal_axis(ax, xlims, ylims, zlims, scale=1.0, dim3=True):
    """Helper function to set equal axis
    
    Args:
        ax (Axes3DSubplot): matplotlib 3D axis, created by `ax = fig.add_subplot(projection='3d')`
        xlims (list): 2-element list containing min and max value of x
        ylims (list): 2-element list containing min and max value of y
        zlims (list): 2-element list containing min and max value of z
        scale (float): scaling factor along x,y,z
        dim3 (bool): whether to also set z-limits (True for 3D plots)
    """
    # compute max required range
    max_range = np.array([max(xlims)-min(xlims), max(ylims)-min(ylims), max(zlims)-min(zlims)]).max() / 2.0
    # compute mid-point along each axis
    mid_x = (max(xlims) + min(xlims)) * 0.5
    mid_y = (max(ylims) + min(ylims)) * 0.5
    mid_z = (max(zlims) + min(zlims)) * 0.5
    # set limits to axis
    if dim3==True:
        ax.set_box_aspect((max_range, max_range, max_range))
    ax.set_xlim(mid_x - max_range*scale, mid_x + max_range*scale)
    ax.set_ylim(mid_y - max_range*scale, mid_y + max_range*scale)
    if dim3==True:
        ax.set_zlim(mid_z - max_range*scale, mid_z + max_range*scale)
    return



"""
Test solving in MEE, forward in time
"""

import numpy as np
from numpy.random import rand
import matplotlib.pyplot as plt
import time 

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))
import pyqlaw

import faulthandler
faulthandler.enable()


def test_solve_mee():
    tstart = time.time()

    # construct problem
    prob = pyqlaw.QLaw(
        integrator="rk4", 
        elements_type="mee_with_a",
        verbosity=2,
        print_frequency=3000,
    )
    # initial and final elements: [a,e,i,RAAN,omega,ta]
    oe0 = pyqlaw.kep2mee_with_a(np.array([1.0, 1e-4, 0.0, 1e-2, 1e-2, 1e-3]))
    oeT = pyqlaw.kep2mee_with_a(np.array([1.8, 0.2, np.pi/6, 1e-2, 1e-2, 1e-3]))
    #oe0 = pyqlaw.kep2mee_with_a(np.array([1.0, 1e-4, 0.0, 0.3, 0.4, 0.1]))
    #oeT = pyqlaw.kep2mee_with_a(np.array([1.52, 0.01, 0.02, 2.3, 0.4, 0.1]))
    print(f"oe0: {oe0}")
    print(f"oeT: {oeT}")
    
    woe = [1.0, 1.0, 1.0, 1.0, 1.0]
    # spacecraft parameters
    mass0 = 1.0
    tmax = 1e-3
    mdot = 1e-3
    tf_max = 10000.0

    # solve forward in time
    t_step = 1.0
    prob.set_problem(oe0, oeT, mass0, tmax, mdot, tf_max, t_step, woe=woe)
    prob.solve(eta_a=0.1, eta_r=0.1)
    prob.pretty_results()
    tend = time.time()

    # plot
    fig1, ax1 = prob.plot_elements_history(to_keplerian=True)
    fig2, ax2 = prob.plot_trajectory_3d(sphere_radius=0.1)
    fig3, ax3 = prob.plot_controls()
    assert prob.converge == True


if __name__=="__main__":
    figs = test_solve_mee()
    plt.show()


# -*- coding: utf-8 -*-

from setuptools import setup
from codecs import open
from os import path
import re

package_name = "pyqlaw"

requires = [
    'sympy>=1.12',
    'numpy>=1.20.4',
    'numba>=0.54.0',
    'matplotlib>=3.8.4',
    'scipy>=1.13.0',
    'tqdm>=4.66.4',
]

root_dir = path.abspath(path.dirname(__file__))

def _requirements():
    return [name.rstrip() for name in open(path.join(root_dir, 'environment.yaml')).readlines()]

def _test_requirements():
    return [name.rstrip() for name in open(path.join(root_dir, 'test-requirements.txt')).readlines()]

with open(path.join(root_dir, "pyqlaw", '__init__.py')) as f:
    init_text = f.read()
    version = re.search(r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    license = re.search(r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author = re.search(r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author_email = re.search(r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)

assert version
assert license
assert author
assert author_email
assert url

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=package_name,
    packages=["pyqlaw"],
    version=version,
    license=license,
    install_requires=requires,
    author=author,
    author_email=author_email,
    url=url,
    description='high-fidelity spaceraft dynamics propagator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='astrodynamics',
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
    ],
)



"""
equations of motion
"""

import numpy as np
from numba import njit


@njit
def eom_kep_gauss(t, state, param):
    """Equations of motion for gauss with keplerian elements"""
    # unpack parameters
    mu, u, psi_c0, psi_c1, psi_c2 = param

    # unpack elements
    sma,ecc,inc,ra,om,ta = state

    p = sma*(1 - ecc**2)
    h = np.sqrt(sma*mu*(1-ecc**2))
    r = h**2/(mu*(1+ecc*np.cos(ta)))

    # Gauss perturbation
    psi = np.array([   
        #[2*sma**2*ecc*np.sin(ta) / h, (2*sma**2*p) / (r*h), 0.0],
        #[p*np.sin(ta) / h, ( (p + r)*np.cos(ta) + r*ecc ) / h, 0.0],
        #[0, 0, r*np.cos(ta+om) / h],
        #[0, 0, r*np.sin(ta+om)/ ( h*np.sin(inc) )],
        #[-p*np.cos(ta) / ( ecc*h ), (p+r)*np.sin(ta) / ( ecc*h ),  -( r*np.sin(ta+om)*np.cos(inc) ) / ( h*np.sin(inc) )],
        [psi_c0[0], psi_c1[0], psi_c2[0]],
        [psi_c0[1], psi_c1[1], psi_c2[1]],
        [psi_c0[2], psi_c1[2], psi_c2[2]],
        [psi_c0[3], psi_c1[3], psi_c2[3]],
        [psi_c0[4], psi_c1[4], psi_c2[4]],
        [p*np.cos(ta) / ( ecc*h ), - (p+r)*np.sin(ta) / ( ecc*h ), 0.0],
    ])

    # combine
    doe = np.dot(psi,u) + np.array([0.0, 0.0, 0.0, 0.0, 0.0, h/r**2])
    return doe


@njit
def eom_mee_gauss(t, state, param):
    """Equations of motion for gauss with MEE"""
    # unpack parameters
    mu, u, psi_c0, psi_c1, psi_c2 = param
    # u = accel*np.array([
    #     np.cos(beta)*np.sin(alpha),
    #     np.cos(beta)*np.cos(alpha),
    #     np.sin(beta),
    # ])

    # unpack elements
    p,f,g,h,k,l = state

    # Gauss perturbation
    sinL = np.sin(l)
    cosL = np.cos(l)
    w = 1 + f*np.cos(l) + g*np.sin(l)
    sqrt_pmu = np.sqrt(p/mu)
    psi = np.array([   
        [0.0, 1/w*sqrt_pmu* 2*p, 0.0],
        [ sqrt_pmu*sinL, sqrt_pmu/w*((w+1)*cosL + f), -g/w*sqrt_pmu*(h*sinL - k*cosL)],
        [-sqrt_pmu*cosL, sqrt_pmu/w*((w+1)*sinL + g),  f/w*sqrt_pmu*(h*sinL - k*cosL)],
        [0.0, 0.0, sqrt_pmu/w * 0.5*(1 + h**2 + k**2)*cosL],
        [0.0, 0.0, sqrt_pmu/w * 0.5*(1 + h**2 + k**2)*sinL],
        [0.0, 0.0, sqrt_pmu/w * (h*sinL - k*cosL)]
    ])

    # combine
    doe = np.dot(psi,u) + np.array([0.0, 0.0, 0.0, 0.0, 0.0, np.sqrt(mu*p)*(w/p)**2])
    return doe


@njit
def eom_mee_with_a_gauss(t, state, param):
    """Equations of motion for gauss with MEE"""
    # unpack parameters
    mu, u, psi_c0, psi_c1, psi_c2 = param

    # unpack elements
    sma,f,g,h,k,l = state

    # compute additional parameters
    ecc = np.sqrt(f**2 + g**2)
    ang_mom = np.sqrt(sma*mu*(1-ecc**2))
    ta = l - np.arctan(g/f)
    r = ang_mom**2/(mu*(1+ecc*np.cos(ta)))
    p = sma*(1 - ecc**2)

    # Gauss perturbation
    sinL = np.sin(l)
    cosL = np.cos(l)
    w = 1 + f*cosL + g*sinL
    sqrt_pmu = np.sqrt(p/mu)
    psi = np.array([   
        # [2*sma**2*ecc*np.sin(ta) / ang_mom, (2*sma**2*p) / (r*ang_mom), 0.0],
        # [ sqrt_pmu*sinL, sqrt_pmu/w*((w+1)*cosL + f), -g/w*sqrt_pmu*(h*sinL - k*cosL)],
        # [-sqrt_pmu*cosL, sqrt_pmu/w*((w+1)*sinL + g),  f/w*sqrt_pmu*(h*sinL - k*cosL)],
        # [0.0, 0.0, sqrt_pmu/w * 0.5*(1 + h**2 + k**2)*cosL],
        # [0.0, 0.0, sqrt_pmu/w * 0.5*(1 + h**2 + k**2)*sinL],
        [psi_c0[0], psi_c1[0], psi_c2[0]],
        [psi_c0[1], psi_c1[1], psi_c2[1]],
        [psi_c0[2], psi_c1[2], psi_c2[2]],
        [psi_c0[3], psi_c1[3], psi_c2[3]],
        [psi_c0[4], psi_c1[4], psi_c2[4]],
        [0.0, 0.0, sqrt_pmu/w * (h*sinL - k*cosL)]
    ])

    # combine
    doe = np.dot(psi,u) + np.array([0.0, 0.0, 0.0, 0.0, 0.0, np.sqrt(mu*p)*(w/p)**2])
    return doe


# @njit
# def eom_gauss_nonspherical_gravity(t, state, p):
#     """Equations of motion for gauss with gravitational perturbation"""
#     # unpack parameters
#     mu, f, alpha, beta = p
#     u = f*np.array([
#         np.cos(beta)*np.sin(alpha),
#         np.cos(beta)*np.cos(alpha),
#         np.sin(beta),
#     ])

#     # unpack elements
#     sma,ecc,inc,ra,om,ta = state

#     p = sma*(1 - ecc**2)
#     h = np.sqrt(sma*mu*(1-ecc**2))
#     r = h**2/(mu*(1+ecc*np.cos(ta)))

#     # Gauss perturbation
#     psi = np.array([   
#         [2*sma**2*ecc*np.sin(ta) / h, (2*sma**2*p) / (r*h), 0.0],
#         [p*np.sin(ta) / h, ( (p + r)*np.cos(ta) + r*ecc ) / h, 0.0],
#         [0, 0, r*np.cos(ta+om) / h],
#         [0, 0, r*np.sin(ta+om)/ ( h*np.sin(inc) )],
#         [-p*np.cos(ta) / ( ecc*h ), (p+r)*np.sin(ta) / ( ecc*h ),  -( r*np.sin(ta+om)*np.cos(inc) ) / ( h*np.sin(inc) )],
#         [p*np.cos(ta) / ( ecc*h ), - (p+r)*np.sin(ta) / ( ecc*h ), 0],
#     ])

#     # combine
#     doe = np.dot(psi,u) + np.array([0.0, 0.0, 0.0, 0.0, 0.0, h/r**2])
#     return doe



"""
main functions for Q-Law transfers
"""

import copy
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d

from ._symbolic import symbolic_qlaw_keplerian, symbolic_qlaw_mee_with_a
from ._lyapunov import lyapunov_control_angles
from ._eom import eom_kep_gauss, eom_mee_gauss, eom_mee_with_a_gauss
from ._integrate import rk4, rkf45
from ._convergence import check_convergence, elements_safety
from ._elements import (
    kep2sv,
    mee_with_a2sv,
    get_orbit_coordinates, 
    mee_with_a2mee,
    mee2mee_with_a,
    mee_with_a2kep,
    ta2ea,
    mee2ea
)
from ._plot_helper import plot_sphere_wireframe, set_equal_axis

import json

class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class QLaw:
    """Object for Q-law based transfer problem.
    The overall procedure for using this class is:
    1. Create object via `prob = QLaw()`
    2. Set problem parameters via `prob.set_problem()`
    3. solve problem via `prob.solve()`

    Exitcodes:
    0 : initial value (problem not yet attempted)
    1 : solved within tolerance
    2 : solved within relaxed tolerance
    -1 : mass is below threshold
    -2 : target elements could not be reached within allocated time
    -3 : thrust angles from feedback control law is nan

    Args:
        mu (float): gravitational parameter, default is 1.0
        rpmin (float): minimum periapsis
        k_petro (float): scalar factor k on minimum periapsis
        m_petro (float): scalar factor m to prevent non-convergence, default is 3.0
        n_petro (float): scalar factor n to prevent non-convergence, default is 4.0
        r_petro (float): scalar factor r to prevent non-convergence, default is 2.0
        b_petro (float): scalar factor b for dot(omega)_xx, default is 0.01
        wp (float): penalty scalar on minimum periapsis, default is 1.0
        elements_type (str): type of elements to define Q-law
        integrator (str): "rk4" or "rkf45"
        verbosity (int): verbosity level for Q-law
        anomaly_grid_size (int): number of evaluation point along orbit for Q-law effectivity
        disable_tqdm (bool): whether to disable progress bar
        tol_oe (np.array or None): tolerance on 5 elements targeted
        oe_min (np.array): minimum values of elements for safe-guarding
        oe_max (np.array): minimum values of elements for safe-guarding
        nan_angles_threshold (int): number of times to ignore `nan` thrust angles
        print_frequency (int): if verbosity >= 2, prints at this frequency
        use_sundman (bool): whether to use Sundman transformation for propagation

    Attributes:
        print_frequency (int): if verbosity >= 2, prints at this frequency
    """
    def __init__(
        self, 
        mu=1.0,
        rpmin=0.1, 
        k_petro=1.0, 
        m_petro=3.0, 
        n_petro=4.0, 
        r_petro=2.0, 
        b_petro=0.01,
        wp=1.0,
        elements_type="mee_with_a",
        integrator="rk4",
        verbosity=1,
        anomaly_grid_size=5,
        tol_oe=None,
        oe_min=None,
        oe_max=None,
        nan_angles_threshold=10,
        print_frequency=200,
        use_sundman=False,
    ):
        """Construct QLaw object"""
        # assertions on inputs
        assert elements_type in ["keplerian", "mee_with_a"], "elements_type must be either 'keplerian' or 'mee_with_a'"
        assert integrator in ["rk4", "rkf45"], "integrator must be either 'rk4' or 'rkf45'"

        # dynamics
        self.mu = mu

        # Q-law parameters
        self.rpmin = rpmin
        self.k_petro = k_petro
        self.m_petro = m_petro
        self.n_petro = n_petro 
        self.r_petro = r_petro
        self.b_petro = b_petro
        self.wp = wp
        self.elements_type = elements_type
        self.integrator = integrator
        self.anomaly_grid_size = anomaly_grid_size

        # settings
        self.verbosity = verbosity

        # tolerance for convergence
        if tol_oe is None:
            self.tol_oe = np.array([1e-3, 1e-3, 1e-3, 1e-3, 1e-3])
        else:
            assert len(tol_oe)==5, "tol_oe must have 5 components"
            self.tol_oe = np.array(tol_oe)
        self.tol_oe_relaxed = 10*self.tol_oe  # relaxed tolerance
        self.exit_at_relaxed = 25

        # orbital elements bounds
        if oe_min is None:
            if self.elements_type == "keplerian":
                self.oe_min = np.array([0.05,1e-4,1e-4,1e-4,1e-4,1e-8])
            elif self.elements_type == "mee_with_a":
                self.oe_min = np.array([0.05, -np.inf, -np.inf, -np.inf, -np.inf])
        else:
            self.oe_min = np.array(oe_min)

        # bounds on elements
        if oe_max is None:
            if self.elements_type == "keplerian":
                self.oe_max = np.array([1e2,0.95,np.pi,np.inf,np.inf,np.inf])
            elif self.elements_type == "mee_with_a":
                self.oe_max = np.array([1e2, 10.0, 10.0, np.inf, np.inf, np.inf])
        else:
            self.oe_max = np.array(oe_max)

        # number of times to accept nan angles
        self.nan_angles_threshold = nan_angles_threshold

        # construct element names, depending on setting
        if self.elements_type == "keplerian":
            self.element_names = ["a", "e", "i", "raan", "omega", "ta"]
            self.eom = eom_kep_gauss
            fun_lyapunov_control, _, _, fun_eval_dqdt = symbolic_qlaw_keplerian()
            self.lyap_fun = fun_lyapunov_control
            self.dqdt_fun = fun_eval_dqdt

        elif self.elements_type == "mee_with_a":
            self.element_names = ["a", "f", "g", "h", "k", "L"]
            self.eom = eom_mee_with_a_gauss
            fun_lyapunov_control, _, _, fun_eval_dqdt = symbolic_qlaw_mee_with_a()
            self.lyap_fun = fun_lyapunov_control
            self.dqdt_fun = fun_eval_dqdt

        # max and min step sizes used with adaptive step integrators
        self.step_min = 1e-4
        self.step_max = 2.0
        self.ode_tol = 1.e-5
        self.use_sundman = use_sundman

        # # duty cycles
        # self.duty_cycle = duty_cycle

        # # battery parameters
        # self.battery_initial = battery_initial
        # self.battery_capacity = battery_capacity
        # self.battery_charge_discharge_rate = battery_charge_discharge_rate
        # self.require_full_recharge = require_full_recharge

        # print frequency
        self.print_frequency = print_frequency  # print at (# of iteration) % self.print_frequency

        # checks
        self.ready = False
        self.converge = False
        self.exitcode = 0
        return
    

    def set_problem(
        self, 
        oe0, 
        oeT, 
        mass0, 
        tmax, 
        mdot, 
        tf_max=100000.0, 
        t_step=0.1,
        mass_min=0.1,
        woe=None,
        duty_cycle=(1e16,0.0),
        battery_initial=1.0,
        battery_capacity=(0.2,1.0),
        battery_charge_discharge_rate=(0.2,0.01),
        require_full_recharge=False,
    ):
        """Set transfer problem
        
        Args:
            oe0 (np.array): initial state, in Keplerian elements (6 components)
            oeT (np.array): final target Keplerian elements (5 components)
            mass0 (float): initial mass
            tmax (float): max thrust
            mdot (float): mass-flow rate
            tf_max (float): max time allocated to transfer
            t_step (float): initial time-step size to be used for integration
            mass_min (float): minimum mass
            woe (np.array): weight on each osculating element
            duty_cycle (tuple): ON and OFF times for duty cycle, default is (1e16, 0.0)
            use_sundman (bool): whether to use Sundman transformation for propagation
            battery_capacity (tuple): min and max battery capacity (min should be DOD)
            battery_charge_discharge_rate (tuple): charge and discharge rate (both positive values)
            require_full_recharge (bool): whether full recharge is required once DOD is reached
        """
        assert len(oe0)==6, "oe6 must have 6 components"
        assert mass_min >= 1e-2, "mass should be above 0.01 to avoid numerical difficulties"
        #assert len(oeT)==5, "oeT must have 5 components"
        # weight parameters
        if woe is None:
            self.woe = np.array([1.0,1.0,1.0,1.0,1.0])
        else:
            assert len(woe)==5, "woe must have 5 components"
            self.woe = np.array(woe)

        # time parameters
        self.tf_max = tf_max
        self.t_step = t_step
        # spacecraft parameters
        self.oe0   = oe0
        if len(oeT) == 6:
            self.oeT = oeT[0:5]
        else:
            self.oeT = oeT
        self.mass0 = mass0
        self.tmax  = tmax
        self.mdot  = mdot
        self.mass_min = mass_min
        self.ready = True               # toggle for solving

        # duty cycles
        self.duty_cycle = duty_cycle

        # battery parameters
        self.battery_initial = battery_initial
        self.battery_capacity = battery_capacity
        self.battery_charge_discharge_rate = battery_charge_discharge_rate
        self.require_full_recharge = require_full_recharge
        return


    def solve(self, eta_a=0.0, eta_r=0.0):
        """Propagate and solve control problem

        Efficiency thresholds `eta_a` and `eta_r` may be given either as constant floats or as callables.
        If providing callables, the required signatures are:

        -   `eta_a(t, oe, mass, battery) -> eta_a_value::float`
        -   `eta_r(t, oe, mass, battery) -> eta_r_value::float`
        
        Args:
            eta_a (float or callable): min absolute effectivity to thrust
            eta_r (float or callable): relative effectivity, `0.0 <= eta_r <= 1.0`
        """
        assert self.ready == True, "Please first call `set_problem()`"

        if isinstance(eta_a, float):
            # `eta_a` is a float
            pass
        elif callable(eta_a):
            # `eta_a` is a callable function
            pass
        else:
            raise ValueError("Invalid type for `eta_a`. Expected float or callable function.")

        # initialize values for propagation
        t_iter = 0.0
        oe_iter = self.oe0
        mass_iter = self.mass0
        battery_iter = self.battery_initial

        # initialize storage
        self.times = [t_iter,]
        self.states = [oe_iter,]
        self.masses = [mass_iter,]
        self.controls = []
        self.etas = []
        self.etas_bounds = []
        self.battery = [self.battery_initial,]
        n_relaxed_cleared = 0
        n_nan_angles = 0

        if self.verbosity >= 2:
            self.disable_tqdm = True
            header = " iter   |  time      |  del1       |  del2       |  del3       |  del4       |  del5       |  el6        |"

        # place-holder for handling duty cycle and battery level
        duty = True
        t_last_ON  = 0.0
        t_last_OFF = 0.0
        charging = False

        # iterate until nmax
        idx = 0
        while abs(self.times[-1]) < abs(self.tf_max):  #for idx in tqdm(range(nmax), disable=self.disable_tqdm, desc="qlaw"):
            # ensure numerical stabilty
            if self.elements_type=="keplerian":
                oe_iter = elements_safety(oe_iter, self.oe_min, self.oe_max)

            # if using Sundman transformation, choose new time-step
            if self.use_sundman:
                # compute eccentric anomaly
                if self.elements_type=="keplerian":
                    ecc_iter = oe_iter[1]
                    E0 = ta2ea(oe_iter[5], oe_iter[1])
                    period = 2*np.pi*np.sqrt(oe_iter[0]**3/self.mu)
                elif self.elements_type=="mee_with_a":
                    ecc_iter = np.sqrt(oe_iter[1]**2 + oe_iter[2]**2)
                    E0 = mee2ea(oe_iter)
                    period = 2*np.pi*np.sqrt(oe_iter[0]**3/self.mu)
                E1 = E0 + self.t_step

                # compute mean anomaly
                M0 = E0 - ecc_iter*np.sin(E0)
                M1 = E1 - ecc_iter*np.sin(E1)
                if M1 > M0:
                    t_step_local = (M1/(2*np.pi) - M0/(2*np.pi)) * period
                else:
                    t_step_local = (M1/(2*np.pi) + 1 - M0/(2*np.pi)) * period
            else:
                t_step_local = self.t_step

            # compute instantaneous acceleration magnitude due to thrust
            accel_thrust = np.sign(t_step_local) * self.tmax/mass_iter

            # evaluate duty cycle
            if ((t_iter - t_last_ON) > self.duty_cycle[0]) and (duty is True):
                duty = False            # turn off duty cycle
                t_last_OFF = t_iter     # latest time when we turn off
            
            if ((t_iter - t_last_OFF) > self.duty_cycle[1]) and (duty is False):
                duty = True             # turn on duty cycle
                t_last_ON = t_iter      # latest time when we turn on

            # check battery state
            if battery_iter - self.battery_charge_discharge_rate[1]*t_step_local < self.battery_capacity[0]:
                duty = False                # turn off
                t_last_OFF = t_iter         # latest time when we turn off
                charging = True

            # check if battery is full
            if (self.require_full_recharge is True) and (charging is True) and\
                (battery_iter < self.battery_capacity[1]):
                duty = False                # turn off

            # initialize efficiency parameters for storage
            val_eta_a, val_eta_r = np.nan, np.nan
            eta_a_current, eta_r_current = np.nan, np.nan
            if duty:
                # evaluate Lyapunov function
                alpha, beta, _, psi = lyapunov_control_angles(
                    fun_lyapunov_control=self.lyap_fun,
                    mu=self.mu, 
                    f=accel_thrust, 
                    oe=oe_iter, 
                    oeT=self.oeT, 
                    rpmin=self.rpmin, 
                    m_petro=self.m_petro, 
                    n_petro=self.n_petro, 
                    r_petro=self.r_petro, 
                    b_petro=self.b_petro, 
                    k_petro=self.k_petro, 
                    wp=self.wp, 
                    woe=self.woe,
                )

                # ensure angles are not nan and otherwise compute thrust vector
                throttle = 1   # initialize
                if np.isnan(alpha) == True or np.isnan(beta) == True:
                    alpha, beta = 0.0, 0.0
                    throttle = 0  # turn off
                    u = np.array([0.0,0.0,0.0])
                    n_nan_angles += 1
                    if n_nan_angles > self.nan_angles_threshold:
                        if self.verbosity > 0:
                            print("Breaking as angles are nan")
                        self.exitcode = -3
                        break
                else:
                    u = accel_thrust*np.array([
                        np.cos(beta)*np.sin(alpha),
                        np.cos(beta)*np.cos(alpha),
                        np.sin(beta),
                    ])

                    # compute effectivity thresholds
                    if isinstance(eta_a, float):        # `eta_a` is a float
                        eta_a_current = eta_a
                    elif callable(eta_a):               # `eta_a` is a callable function
                        eta_a_current = eta_a(t_iter, oe_iter, mass_iter, battery_iter)

                    if isinstance(eta_r, float):        # `eta_r` is a float
                        eta_r_current = eta_r
                    elif callable(eta_r):               # `eta_r` is a callable function
                        eta_r_current = eta_r(t_iter, oe_iter, mass_iter, battery_iter)

                    # check effectivity to decide whether to thrust or coast
                    if eta_r_current > 0 or eta_a_current > 0:
                        qdot_current = self.dqdt_fun(
                            self.mu, 
                            accel_thrust, 
                            oe_iter, 
                            self.oeT, 
                            self.rpmin, self.m_petro, self.n_petro, 
                            self.r_petro, self.b_petro, self.k_petro, 
                            self.wp, self.woe
                        )
                        qdot_min, qdot_max = self.evaluate_osculating_qdot(
                            oe_iter, accel_thrust
                        )
                        val_eta_a = qdot_current/qdot_min
                        val_eta_r = (qdot_current - qdot_max)/(qdot_min - qdot_max)
                        # turn thrust off if below threshold
                        if val_eta_a < eta_a_current or val_eta_r < eta_r_current:
                            throttle = 0  # turn off
                            u = np.zeros((3,))
            else:
                u = np.zeros((3,))
                throttle = 0  # turn off
                # evaluate Lyapunov function just for psi (FIXME)
                _, _, _, psi = lyapunov_control_angles(
                    fun_lyapunov_control=self.lyap_fun,
                    mu=self.mu, 
                    f=accel_thrust, 
                    oe=oe_iter, 
                    oeT=self.oeT, 
                    rpmin=self.rpmin, 
                    m_petro=self.m_petro, 
                    n_petro=self.n_petro, 
                    r_petro=self.r_petro, 
                    b_petro=self.b_petro, 
                    k_petro=self.k_petro, 
                    wp=self.wp, 
                    woe=self.woe,
                )
                #print(f"throttle = {throttle}")

            # ODE parameters
            ode_params = (self.mu, u, psi[0], psi[1], psi[2])  # control fixed in RTN for step
            if self.integrator == "rk4":
                oe_next = rk4(
                    self.eom, 
                    t_iter,
                    t_step_local,
                    oe_iter,
                    ode_params,
                )
                t_iter += t_step_local  # update time
                if throttle == 1:
                    mass_iter -= self.mdot*t_step_local  # update mass
                oe_iter = oe_next

            elif self.integrator == "rkf45":
                oe_next, h_next = rkf45(
                    self.eom, 
                    t_iter,
                    t_step_local,
                    oe_iter,
                    ode_params,
                    self.ode_tol,
                )
                t_iter += t_step_local  # update time
                if throttle == 1:
                    mass_iter -= self.mdot*t_step_local  # update mass
                oe_iter = oe_next
                #print(f"h_nect: {h_next}")
                t_step_local = max(self.step_min, min(self.step_max,h_next))
            else:
                raise ValueError("integrator name invalid!")
                
            # check convergence
            if check_convergence(oe_next, self.oeT, self.woe, self.tol_oe) == True:
                self.exitcode = 1
                self.converge = True
                break

            # check relaxed condition
            if check_convergence(oe_next, self.oeT, self.woe, self.tol_oe_relaxed) == True:
                n_relaxed_cleared += 1
                if n_relaxed_cleared >= self.exit_at_relaxed:
                    self.exitcode = 2
                    self.converge = True
                    break
            
            # print progress
            if self.verbosity >= 2 and np.mod(idx,self.print_frequency)==0:
                if np.mod(idx, 20*self.print_frequency) == 0:
                    print("\n" + header)
                #t_fraction = t_iter/self.tf_max
                print(f" {idx:6.0f} | {t_iter: 1.3e} | {oe_next[0]-self.oeT[0]: 1.4e} | {oe_next[1]-self.oeT[1]: 1.4e} | {oe_next[2]-self.oeT[2]: 1.4e} | {oe_next[3]-self.oeT[3]: 1.4e} | {oe_next[4]-self.oeT[4]: 1.4e} | {oe_next[5]: 1.4e} |")

            # check if mass is below threshold
            if mass_iter <= self.mass_min:
                if self.verbosity > 0:
                    print("Breaking as mass is now under mass_min")
                self.exitcode = -1
                break

            # store
            self.times.append(t_iter)
            self.states.append(oe_iter)
            self.masses.append(mass_iter)
            self.controls.append([alpha, beta, throttle])
            self.etas.append([val_eta_a, val_eta_r])
            self.etas_bounds.append([eta_a_current, eta_r_current])

            # update battery
            if duty:
                battery_iter = np.clip(battery_iter-self.battery_charge_discharge_rate[1]*t_step_local,
                                       self.battery_capacity[0], self.battery_capacity[1])
            else:
                battery_iter = np.clip(battery_iter+self.battery_charge_discharge_rate[0]*t_step_local,
                                       self.battery_capacity[0], self.battery_capacity[1])
                if battery_iter == self.battery_capacity[1]:
                    charging = False        # turn OFF charging mode
            self.battery.append(battery_iter)

            # index update
            idx += 1

        if self.converge == False:
            if self.verbosity > 0:
                print("Could not arrive to target elements within time")
            self.exitcode = -2
        else:
            if self.verbosity > 0:
                print("Target elements successfully reached!")
        return

    
    def evaluate_osculating_qdot(self, oe, accel_thrust):
        """Evaluate Qdot over the entire orbit
        
        Args:
            oe (np.array): current osculating elements
            accel_thrust (float): acceleration magnitude, tmax/mass

        Returns:
            (tuple): min and max Qdot
        """
        # evaluate qdot at current as well as for all anomalies
        eval_pts = np.linspace(oe[5], oe[5]+2*np.pi, self.anomaly_grid_size+1)[1:]
        # storage FIXME
        qdot_list = []
        for anomaly in eval_pts:
            # construct element
            oe_test = np.array([oe[0], oe[1], oe[2], oe[3], oe[4], anomaly])
            # # evaluate Lyapunov function
            # alpha, beta, _, psi_test = lyapunov_control_angles(
            #     fun_lyapunov_control=self.lyap_fun,
            #     mu=self.mu, 
            #     f=accel_thrust, 
            #     oe=oe_test, 
            #     oeT=self.oeT, 
            #     rpmin=self.rpmin, 
            #     m_petro=self.m_petro, 
            #     n_petro=self.n_petro, 
            #     r_petro=self.r_petro, 
            #     b_petro=self.b_petro, 
            #     k_petro=self.k_petro, 
            #     wp=self.wp, 
            #     woe=self.woe,
            # )
            # u_test = accel_thrust*np.array([
            #     np.cos(beta)*np.sin(alpha),
            #     np.cos(beta)*np.cos(alpha),
            #     np.sin(beta),
            # ])

            # evaluate qdot
            qdot_list.append(
                self.dqdt_fun(
                    self.mu, 
                    accel_thrust, 
                    oe_test, 
                    self.oeT, 
                    self.rpmin, self.m_petro, self.n_petro, 
                    self.r_petro, self.b_petro, self.k_petro, 
                    self.wp, self.woe
                )
            )
            #eval_qdot(psi_test, u_test)
        return min(qdot_list), max(qdot_list)


    def plot_elements_history(
        self,
        figsize = (10,6),
        to_keplerian = False,
        TU = 1.0,
        time_unit_name = "TU",
        plot_mass = True,
        plot_periapsis = False,
    ):
        """Plot elements time history
        
        Args:
            figsize (tuple): figure size
            to_keplerian (bool): whether to convert to Keplerian elements
            TU (float): time unit
            time_unit_name (str): name of time unit
            plot_mass (bool): whether to plot mass or anomaly
            plot_periapsis (bool): whether to overlay periapsis history on top of first plot
        
        Returns:
            (tuple): figure and axis objects
        """
        oes = np.zeros((6,len(self.times)))
        fig, axs = plt.subplots(2,3,figsize=figsize)
        for idx in range(len(self.times)):
            if to_keplerian == True and self.elements_type=="mee_with_a":
                oes[:,idx] = mee_with_a2kep(self.states[idx])
                labels = ["a", "e", "i", "raan", "om", "ta"]
                multipliers = [1,1,180/np.pi,180/np.pi,180/np.pi,180/np.pi]
                oe_T = mee_with_a2kep(np.concatenate((self.oeT,[0.0])))[0:5]
                show_band = False
            else:
                oes[:,idx] = self.states[idx]
                labels = self.element_names
                multipliers = [1,1,1,1,1,1]
                oe_T = self.oeT
                show_band = True

        for idx,ax in enumerate(axs.flatten()[:5]):
            # targeted elements
            if show_band:
                ax.fill_between(
                    np.array(self.times)*TU, 
                    (oe_T[idx]-self.tol_oe[idx])*multipliers[idx], 
                    (oe_T[idx]+self.tol_oe[idx])*multipliers[idx], 
                    color='red', alpha=0.25
                )
            else:
                ax.axhline(oe_T[idx]*multipliers[idx], color='r', linestyle='--')

            # state history
            ax.plot(np.array(self.times)*TU, oes[idx,:]*multipliers[idx], label=labels[idx])
            ax.set(xlabel=f"Time, {time_unit_name}", ylabel=labels[idx])

        # overlay periapsis
        if plot_periapsis:
            axs[0,0].plot(np.array(self.times)*TU, oes[0,:]*(1 - oes[1,:]))

        # plot mass or TA
        if plot_mass:
            axs[1,2].plot(np.array(self.times)*TU, self.masses)
            axs[1,2].set(xlabel=f"Time, {time_unit_name}", ylabel="Mass")
        else:
            axs[1,2].plot(np.array(self.times)*TU, (oes[5,:] % (2*np.pi))*180/np.pi, label=labels[5])
            axs[1,2].set(xlabel=f"Time, {time_unit_name}", ylabel=labels[5])
        plt.tight_layout()
        return fig, ax


    def plot_battery_history(
        self,
        figsize=(9,5),
        TU=1.0,
        BU=1.0,
        time_unit_name="TU",
        battery_unit_name="BU"
    ):
        """Plot battery history
        
        Args:
            figsize (tuple): figure size
            TU (float): time unit
            BU (float): battery unit
            time_unit_name (str): name of time unit
            battery_unit_name (str): name of battery unit
        
        Returns:
            (tuple): figure and axis objects
        """
        fig, ax = plt.subplots(1,1,figsize=figsize)
        ax.plot(np.array(self.times)*TU, np.array(self.battery)*BU, color='k')
        ax.axhline(np.array(self.battery_capacity[0])*BU, color='r', linestyle='--', label="min capacity")
        ax.axhline(np.array(self.battery_capacity[1])*BU, color='g', linestyle='--', label="max capacity")
        ax.set(xlabel=f"Time, {time_unit_name}", ylabel=f"Battery, {battery_unit_name}")
        return fig, ax


    def plot_controls(self, figsize=(9,5), TU=1.0, time_unit_name="TU"):
        """Plot control time history
        
        Args:
            figsize (tuple): figure size
            TU (float): time unit
            time_unit_name (str): name of time unit
        
        Returns:
            (tuple): figure and axis objects
        """
        alphas, betas, throttles = [], [], []
        for control in self.controls:
            alphas.append(control[0])
            betas.append(control[1])
            throttles.append(control[2])
        fig, ax = plt.subplots(1,1,figsize=figsize)
        ax.step(np.array(self.times[0:-1])*TU, np.array(alphas)*180/np.pi, where='pre', marker='o', markersize=2, label="alpha")
        ax.step(np.array(self.times[0:-1])*TU, np.array(betas)*180/np.pi, where='pre', marker='o', markersize=2, label="beta")
        ax.step(np.array(self.times[0:-1])*TU, np.array(throttles)*100, where='pre', marker='o', markersize=2, label="throttle, %")
        ax.set(xlabel=f"Time, {time_unit_name}", ylabel="Control angles and throttle")
        ax.legend()
        return fig, ax


    def interpolate_states(self, kind="quadratic"):
        """Create interpolation states"""
        # prepare states matrix
        state_matrix = np.zeros((6,len(self.states)))
        for idx,state in enumerate(self.states):
            state_matrix[:,idx] = state
        f_a = interp1d(self.times, state_matrix[0,:], kind=kind)
        f_e = interp1d(self.times, state_matrix[1,:], kind=kind)
        f_i = interp1d(self.times, state_matrix[2,:], kind=kind)
        f_r = interp1d(self.times, state_matrix[3,:], kind=kind)
        f_o = interp1d(self.times, state_matrix[4,:], kind=kind)
        f_t = interp1d(self.times, state_matrix[5,:], kind=kind)
        return (f_a, f_e, f_i, f_r, f_o, f_t)


    def get_cartesian_history(self, interpolate=False, steps=None):
        """Get Cartesian history of states"""
        if interpolate:
            # interpolate orbital elements
            f_a, f_e, f_i, f_r, f_o, f_t = self.interpolate_states()
            if steps is None:
                steps = min(8000, abs(int(round(self.times[-1]/0.1))))
                print(f"Using {steps} steps for evaluation")
            t_evals = np.linspace(self.times[0], self.times[-1], steps)
            cart = np.zeros((6,steps))
            for idx, t in enumerate(t_evals):
                if self.elements_type=="keplerian":
                    cart[:,idx] = kep2sv([f_a(t), f_e(t), f_i(t), f_r(t), f_o(t), f_t(t)], self.mu)
                elif self.elements_type=="mee_with_a":
                    cart[:,idx] = mee_with_a2sv([f_a(t), f_e(t), f_i(t), f_r(t), f_o(t), f_t(t)], self.mu)
        else:
            cart = np.zeros((6,len(self.times)))
            for idx in range(len(self.times)):
                if self.elements_type=="keplerian":
                    cart[:,idx] = kep2sv(self.states[idx], self.mu)
                elif self.elements_type=="mee_with_a":
                    cart[:,idx] = mee_with_a2sv(np.array(self.states[idx]), self.mu)
        return cart


    def plot_trajectory_2d(
        self, 
        figsize=(6,6),
        interpolate=False, 
        steps=None, 
    ):
        """Plot trajectory in xy-plane"""
        # get cartesian history
        cart = self.get_cartesian_history(interpolate, steps)

        fig, ax = plt.subplots(1,1,figsize=figsize)
        # plot initial and final orbit
        if self.elements_type == "keplerian":
            coord_orb0 = get_orbit_coordinates(self.oe0, self.mu)
            coord_orbT = get_orbit_coordinates(np.concatenate((self.oeT,[0.0])), self.mu)
        elif self.elements_type == "mee_with_a":
            coord_orb0 = get_orbit_coordinates(mee_with_a2kep(self.oe0), self.mu)
            coord_orbT = get_orbit_coordinates(np.concatenate((mee_with_a2kep(self.oeT)[0:5],[0.0])), self.mu)
        ax.plot(coord_orb0[0,:], coord_orb0[1,:], label="Initial", c="darkblue")
        ax.plot(coord_orbT[0,:], coord_orbT[1,:], label="Final", c="forestgreen")

        # plot transfer
        ax.plot(cart[0,:], cart[1,:], label="transfer", c="crimson", lw=0.4)
        ax.scatter(cart[0,0], cart[1,0], label=None, c="crimson", marker="x")
        ax.scatter(cart[0,-1], cart[1,-1], label=None, c="crimson", marker="o")
        ax.set_aspect('equal')
        return fig, ax


    def plot_trajectory_3d(
        self, 
        figsize=(6,6), 
        interpolate=True, 
        steps=None, 
        plot_sphere=True,
        sphere_radius=0.35,
        scale=1.02,
        lw = 0.4
    ):
        """Plot trajectory in xyz"""
        # get cartesian history
        cart = self.get_cartesian_history(interpolate, steps)

        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(projection='3d')
        # plot initial and final orbit
        if self.elements_type == "keplerian":
            coord_orb0 = get_orbit_coordinates(self.oe0, self.mu)
            coord_orbT = get_orbit_coordinates(np.concatenate((self.oeT,[0.0])), self.mu)
        elif self.elements_type == "mee_with_a":
            coord_orb0 = get_orbit_coordinates(mee_with_a2kep(self.oe0), self.mu)
            coord_orbT = get_orbit_coordinates(
                mee_with_a2kep(np.concatenate((self.oeT, [0.0]))), 
                self.mu
            )

        # plot transfer
        ax.plot(cart[0,:], cart[1,:], cart[2,:], label="transfer", c="crimson", lw=lw)
        ax.scatter(cart[0,0], cart[1,0], cart[2,0], label=None, c="crimson", marker="x")
        ax.scatter(cart[0,-1], cart[1,-1], cart[2,-1], label=None, c="crimson", marker="o")

        ax.plot(coord_orb0[0,:], coord_orb0[1,:], coord_orb0[2,:], label="Initial", c="darkblue")
        ax.plot(coord_orbT[0,:], coord_orbT[1,:], coord_orbT[2,:], label="Final", c="forestgreen")

        # plot center body
        if plot_sphere:
            plot_sphere_wireframe(ax, sphere_radius)
            xlims = [min(cart[0,:]), max(cart[0,:])]
            ylims = [min(cart[1,:]), max(cart[1,:])]
            zlims = [min(cart[2,:]), max(cart[2,:])]
            set_equal_axis(ax, xlims, ylims, zlims, scale=scale)

        #ax.set_aspect('equal')
        ax.set(xlabel="x", ylabel="y", zlabel="z")
        return fig, ax


    def plot_efficiency(self, figsize=(9,6), TU=1.0, time_unit_name="TU"):
        """Plot efficiency
        
        Args:
            figsize (tuple): figure size
            TU (float): time unit
            time_unit_name (str): name of time unit
        
        Returns:
            (tuple): figure and axis objects
        """
        fig, axs = plt.subplots(2,1,figsize=figsize)
        axs[0].plot(np.array(self.times[0:-1])*TU, np.array(self.etas)[:,0], label="eta_a")
        axs[0].plot(np.array(self.times[0:-1])*TU, np.array(self.etas_bounds)[:,0], label="eta_a_min", linestyle='-', color='red')
        axs[0].set(xlabel=f"Time, {time_unit_name}", ylabel="Absolute efficiency")

        axs[1].plot(np.array(self.times[0:-1])*TU, np.array(self.etas)[:,1], label="eta_r")
        axs[1].plot(np.array(self.times[0:-1])*TU, np.array(self.etas_bounds)[:,1], label="eta_r_min", linestyle='-', color='red')
        axs[1].set(xlabel=f"Time, {time_unit_name}", ylabel="Relative efficiency")
        for ax in axs:
            ax.legend()
        plt.tight_layout()
        return fig, ax


    def pretty(self):
        """Pretty print"""
        print(f"Transfer:")
        print(f"  {self.element_names[0]}  : {self.oe0[0]:1.4e} -> {self.oeT[0]:1.4e} (weight: {self.woe[0]:2.2f})")
        print(f"  {self.element_names[1]}  : {self.oe0[1]:1.4e} -> {self.oeT[1]:1.4e} (weight: {self.woe[1]:2.2f})")
        print(f"  {self.element_names[2]}  : {self.oe0[2]:1.4e} -> {self.oeT[2]:1.4e} (weight: {self.woe[2]:2.2f})")
        print(f"  {self.element_names[3]}  : {self.oe0[3]:1.4e} -> {self.oeT[3]:1.4e} (weight: {self.woe[3]:2.2f})")
        print(f"  {self.element_names[4]}  : {self.oe0[4]:1.4e} -> {self.oeT[4]:1.4e} (weight: {self.woe[4]:2.2f})")
        return


    def pretty_results(self):
        """Pretty print results"""
        print(f"Exit code : {self.exitcode}")
        print(f"Converge  : {self.converge}")
        print(f"Final state:")
        print(f"  {self.element_names[0]}  : {self.states[-1][0]:1.4e} (error: {abs(self.states[-1][0]-self.oeT[0]):1.4e})")
        print(f"  {self.element_names[1]}  : {self.states[-1][1]:1.4e} (error: {abs(self.states[-1][1]-self.oeT[1]):1.4e})")
        print(f"  {self.element_names[2]}  : {self.states[-1][2]:1.4e} (error: {abs(self.states[-1][2]-self.oeT[2]):1.4e})")
        print(f"  {self.element_names[3]}  : {self.states[-1][3]:1.4e} (error: {abs(self.states[-1][3]-self.oeT[3]):1.4e})")
        print(f"  {self.element_names[4]}  : {self.states[-1][4]:1.4e} (error: {abs(self.states[-1][4]-self.oeT[4]):1.4e})")
        print(f"Transfer time : {self.times[-1]}")
        print(f"Final mass    : {self.masses[-1]}")
        return


    def pretty_settings(self):
        """Pretty print settings"""
        print(f"Element type  : {self.elements_type}")
        print(f"Element names : {self.element_names}")
        print(f"Integrator    : {self.integrator}")
        print(f"Tolerance     : {self.tol_oe}")
        print(f"Relaxed tolerance : {self.tol_oe_relaxed}")
        print(f"Exit at relaxed   : {self.exit_at_relaxed}")
        return

    def save_to_dict(self, filepath, canonical_units=None, save_control_angles = False):
        """Export result into a dictionary, saved as json if filepath is provided
        
        Args:
            filepath (str): filepath to json filename to save the dictionary
        """
        initial_guess = {
            "t0": 0.0,
            "tf": self.times[-1],
            "times": self.times,
            "canonical_units": canonical_units,
            "states": [list(mee_with_a2mee(oe))+[m] for (oe,m) in zip(self.states, self.masses)],
        }
        _controls = copy.deepcopy(self.controls)
        _controls.append([0,0,0])   # final step to match time-steps
        _controls = np.array(_controls)
        if save_control_angles:
            initial_guess["controls"] = _controls
        else:
            # convert to RTN unit vectors
            alphas    = _controls[:,0]
            betas     = _controls[:,1]
            taus      = _controls[:,2]
            uR = np.cos(betas) * np.sin(alphas)
            uT = np.cos(betas) * np.cos(alphas)
            uN = np.sin(betas)
            initial_guess["controls"] = np.stack((uR,uT,uN,taus)).T

        if filepath is not None:
            #dumped = json.dumps(initial_guess, cls=NumpyEncoder)
            with open(filepath, 'w') as f:
                json.dump(initial_guess, f, cls=NumpyEncoder, indent=4)
        return initial_guess



"""
Transformation functions
"""

from numba import njit
import numpy as np

@njit
def rotmat1(phi):
	return np.array([ [1.0, 0.0, 0.0],
					  [0.0, np.cos(phi), np.sin(phi)],
					  [0.0, -np.sin(phi), np.cos(phi)] ])

@njit
def rotmat2(phi):
	return np.array([ [np.cos(phi), 0.0, -np.sin(phi)],
					  [0.0, 1.0, 0.0],
					  [np.sin(phi), 0.0, np.cos(phi)] ])

@njit
def rotmat3(phi):
	return np.array([ [ np.cos(phi), np.sin(phi), 0.0],
					  [-np.sin(phi), np.cos(phi), 0.0],
					  [0.0, 0.0, 1.0] ])
